package com.sillycycle.bagleyd.panex;

/*
 * @(#)PanexCanvas.java
 *
 * Copyright 1994 - 2022  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Graphics;
import java.awt.Point;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.IOException;
import javax.swing.JFrame;

import com.sillycycle.bagleyd.panex.model.PanexFormat;
import com.sillycycle.bagleyd.panex.model.PanexTile;
import com.sillycycle.bagleyd.panex.view.PanexDraw;
import com.sillycycle.bagleyd.panex.view.PanexGeometry;

/**
 * The <code>PanexCanvas</code> class does all the graphics for
 * <code>PanexFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

/*
References:
  Sliding Piece Puzzles (Recreations in Mathematics Series) by
  L. E. Horden, Oxford University Press 1986, pp 144, 145.

  Puzzles Old & New (How to Make and Solve Them) by Jerry
  Slocum & Jack Botermans, University of Washington Press,
  Seattle, 1987, p 135.

  Nesting Puzzles Part 1: Moving oriental towers by Vladimir
  Dubrovsky, Quantum January/February 1996 pp 55-57, 50-51.
*/

public class PanexCanvas extends Canvas {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");

	static final int TOP = 0;
	static final int RIGHT = 1;
	static final int BOTTOM = 2;
	static final int LEFT = 3;
	static final int UP = 1;
	static final int DOWN = 0;
	static final int NORMAL = 1;
	static final int DOUBLE = 2;
	static final int INSTANT = 3;
	static final int BASE = 10;
	static final String DATAFILE = "panex.dat";
	static final String SYMBOL = ":";

	protected transient PanexFormat model = null;
	protected transient PanexGeometry geo = null;
	protected transient PanexDraw aDraw = null;
	Point[][]   startLoc = null;
	int	 currentStack = -1, nextStack = -1, previousStack = -1;
	int	 currentPosition, nextPosition;
	boolean     started, cheat;
	//boolean     mono, reverse;
	//int	 delay, numSlices;

	JFrame      frame = null;
	boolean     firstPaint = false, resizePaint = false;
	boolean     movePaint = false, scanPaint = false, stackPaint = false;
	boolean     framePaint = false, resetPaint = false;
	boolean     selectPaint = false, motionPaint = false;
	boolean     releasePaint = false;
	boolean     mouseDown = false;
	boolean     focus = true;
	transient PanexStack  undo = null, redo = null;
	transient PanexSolve  solve = null;
	String      stringSave = null;
	String[]    token = null;

	public PanexFormat getModel() {
		return model;
	}

	void checkTiles() {
		model.checkFormat();
		geo.checkGeometry();
	}

	/* Only applies to Panex, halfway done check */
	boolean checkMiddle() {
		int stack, loc, number;
		Integer color;
		Integer[] middlePositions = new Integer[PanexFormat.MAX_COLUMNS];

		middlePositions[0] = null;
		middlePositions[1] = 0;
		middlePositions[2] = 1;
		for (stack = 0; stack < PanexFormat.MAX_COLUMNS; stack++) {
			if ((color = middlePositions[stack]) != null) {
				for (loc = 0; loc < model.getTiles(); loc++) {
					number = loc - 1;
					PanexTile tile = model.getColumn(stack).getTileAtPosition(loc + model.getBlanks());
					if (tile == null ||
							tile.getColor() != color || tile.getNumber() != number) {
						return false;
					}
				}
			}
		}
		return true;
	}

	public boolean checkSolved() {
		int stack, loc, number;
		Integer color;

		for (stack = 0; stack < PanexFormat.MAX_COLUMNS; stack++) {
			if ((color = model.getFinalPosition(stack)) >= 0) {
				for (loc = 0; loc < model.getTiles(); loc++) {
					number = loc;
					PanexTile tile = model.getColumn(stack).getTileAtPosition(loc + model.getBlanks());
					if (tile == null ||
							tile.getColor() != color || tile.getNumber() != number) {
						return false;
					}
				}
			}
		}
		started = false;
		return true;
	}

	void animateSlide(int currentColumn, int currentDepth,
			int nextColumn, int nextDepth, int fast) {
		Graphics g = getGraphics();
		int diffStack = nextColumn - currentColumn;
		int diffPosition = nextDepth - currentDepth;
		int dir = 0, spaces = 0, inc;
		int gapI = 0, moveI = 0, space;
		int dx, dy;
		int ix = 0, iy = 0;

		if (g == null)
			return;
		try {
			if (diffPosition > 0) {
				dir = BOTTOM;
				spaces = diffPosition;
				moveI = geo.getPos().y;
			} else if (diffPosition < 0) {
				dir = TOP;
				spaces = -diffPosition;
				moveI = geo.getPos().y;
			} else if (diffStack > 0) {
				dir = RIGHT;
				spaces = diffStack;
				moveI = geo.getPos().x;
			} else if (diffStack < 0) {
				dir = LEFT;
				spaces = -diffStack;
				moveI = geo.getPos().x;
			}
			space = spaces * moveI;
			if ((dir == RIGHT) || (dir == LEFT)) {
				gapI = geo.getPos().x * fast / geo.getNumberSlices();
			} else if ((dir == TOP) || (dir == BOTTOM)) {
				gapI = geo.getPos().y * fast / geo.getNumberSlices();
			}
			if (gapI == 0)
				gapI++;
			for (inc = 0; inc < space + gapI; inc += gapI) {
				if (inc > space) {
					inc = space;
				}
				/* Calculate deltas */
				dx = geo.cartesianX(currentColumn);
				dy = geo.cartesianY(currentDepth);
				if ((dir == RIGHT) || (dir == LEFT)) {
					ix = ((dir == RIGHT) ? inc : -inc);
					iy = 0;
					aDraw.drawSlot(g, currentColumn,
						dx, dy + geo.getTileSize().y,
						geo.getTileSize().x + 1, 1);
				} else if ((dir == TOP) || (dir == BOTTOM)) {
					ix = 0;
					iy = ((dir == BOTTOM) ? inc : -inc);
				}
				aDraw.drawTile(g, currentColumn, currentDepth,
					false, 0, ix, iy);
				/* Erase old slivers */
				ix += dx;
				iy += dy;
				if (inc != 0)
				switch (dir) {
				case TOP:
					aDraw.drawSlot(g, currentColumn,
						ix, iy + geo.getTileSize().y,
						geo.getTileSize().x + 1, gapI);
					try {
						Thread.sleep(geo.getDelay() / fast);
					} catch (InterruptedException e) {
						// ignore
					}
					break;
				case RIGHT:
					aDraw.drawSlot(g, currentColumn,
						ix - gapI, iy,
						gapI, geo.getTileSize().y + 1);
					try {
						Thread.sleep(3 * geo.getDelay() / fast);
					} catch (InterruptedException e) {
						// ignore
					}
					break;
				case BOTTOM:
					aDraw.drawSlot(g, currentColumn,
						ix, iy - gapI,
						geo.getTileSize().x + 1, gapI);
					try {
						Thread.sleep(geo.getDelay() / fast);
					} catch (InterruptedException e) {
						// ignore
					}
					break;
				case LEFT:
					aDraw.drawSlot(g, currentColumn,
						ix + geo.getTileSize().x, iy,
						gapI + 1, geo.getTileSize().y + 1);
					try {
						Thread.sleep(3 * geo.getDelay() / fast);
					} catch (InterruptedException e) {
						// ignore
					}
					break;
				default:
				}
			}
		} finally {
			g.dispose();
		}
	}

	//TODO remove
	Integer topOfStack(int stack, int start) {
		if (stack < 0 || start < 0) {
			System.err.println("topOfStack: stack = " + stack +
				", start = " + start);
		}
		return model.getColumn(stack).getTopTilePosition(start);
	}

	boolean jumpMove(int fromStack, int toStack) {
		return ((model.getMode() == PanexFormat.Modes.ALGORITHME.ordinal()) &&
			(model.getColumn(1).getTileAtPosition(0) != null) &&
			((toStack == 0 && fromStack == 2) ||
			(toStack == 2 && fromStack == 0)));
	}

	int requestMove(int fromStack, int fromPosition, int toStack,
			boolean limbo) {
		int bottom, depth;

		/* Do not have to check above stack since it is the top one */
		if (toStack > fromStack) {
			if (!jumpMove(fromStack, toStack)) {
				for (bottom = fromStack + 1; bottom <= toStack; bottom++) {
					if (model.getColumn(bottom).getTileAtPosition(0).getColor() >= 0) {
						return -1;
					}
				}
			}
			bottom = topOfStack(toStack, 0);
		} else if (toStack < fromStack) {
			if (!jumpMove(fromStack, toStack)) {
				for (bottom = fromStack - 1; bottom >= toStack; bottom--) {
					if (model.getColumn(bottom).getTileAtPosition(0).getColor() >= 0) {
						return -1;
					}
				}
			}
			bottom = topOfStack(toStack, 0);
		} else {
			/* Ruled out toStack == fromStack, except for motion */
			bottom = topOfStack(toStack, 1);
		}
		depth = (bottom == -1) ? model.getSlots() - 1 : bottom - 1;
		PanexTile movingTile =
			model.getColumn(fromStack).getTileAtPosition(fromPosition);
		if (movingTile == null) {
			System.err.println("requestMove: movingTile is null, "
				+ ", column = " + fromStack + ", depth = " + fromPosition);
			return -1;
		}
		if (model.getMode() == PanexFormat.Modes.PANEX.ordinal()) {
			if (depth > movingTile.getNumber() + 1) {
				return (movingTile.getNumber() + 1);
			}
			return depth;
		} else {
			if (depth == model.getSlots() - 1 || limbo) {
				return depth;
			}
			PanexTile restingTile = model.getColumn(toStack).getTileAtPosition(depth + 1);
			if (restingTile == null) {
				System.err.println("requestMove: restingTile is null, "
						+ ", column = " + toStack + ", depth = " + depth + 1);
				return -1;
			}
			if (movingTile.getColor() != -1 && restingTile.getColor() != -1 &&
					restingTile.getNumber() >= movingTile.getNumber()) {
				return depth;
			}
			return -2;
		}
	}

	void virtualMove(int currentColumn, int currentDepth,
			int nextColumn, int nextDepth) {
		if (currentColumn == -1 || nextColumn == -1) {
			System.err.println("virtualMove: currentStack = " + currentColumn +
				", nextStack = " + nextColumn);
		}
		int color, number;

		PanexTile currentTile = model.getColumn(currentColumn).getTileAtPosition(currentDepth);
		if (currentTile == null) {
			System.err.println("virtualMove: currentTile is null"
				+ ", column = " + currentColumn + ", depth = " + currentDepth);
			return;
		}
		color = currentTile.getColor();
		number = currentTile.getNumber();
		if (color < 0 || number < 0) {
			System.err.println("virtualMove: color = " + color
				+ ", number = " + number);
		}
		model.getColumn(nextColumn).setTileAtPosition(nextDepth, currentTile);
		model.setColumnOfTile(color, number, nextColumn);
		model.setDepthOfTile(color, number, nextDepth);
		model.getColumn(currentColumn).setTileAtPosition(currentDepth, new PanexTile(-1, -1));
	}

	void discreteMoves(int currentColumn, int currentDepth,
			int nextColumn, int nextDepth) {
		aDraw.drawTile(currentColumn, currentDepth, true, 0, 0, 0);
		virtualMove(currentColumn, currentDepth,
			nextColumn, nextDepth);
		aDraw.drawTile(nextColumn, nextDepth, false, 0, 0, 0);
		try {
			Thread.sleep(PanexGeometry.MULT_DELAY * geo.getDelay() /
				(model.getTiles() + PanexFormat.MAX_COLUMNS - 1));
		} catch (InterruptedException e) {
			// ignore
		}
	}

	void slideTile(int fromStack, int fromPosition,
			int toStack, int toPosition, boolean jump, int fast) {
		int currentColumn = fromStack, currentDepth = fromPosition;
		int nextColumn = fromStack, nextDepth = fromPosition;

		if (currentDepth > 0) {
			if (fast != INSTANT && geo.getDelay() > 0) {
				nextDepth = 0;
				animateSlide(currentColumn, currentDepth,
					nextColumn, nextDepth, fast);
				virtualMove(currentColumn, currentDepth,
					nextColumn, nextDepth);
				currentDepth = nextDepth;
			} else {
				while (currentDepth > 0) {
					nextDepth = currentDepth - 1;
					discreteMoves(currentColumn, currentDepth,
						nextColumn, nextDepth);
					currentDepth = nextDepth;
				}
			}
			if (((PanexFrame) frame).getToggleSound())
				((PanexFrame) frame).playMoveAudio();
		}
		if (currentColumn != toStack) {
			if (fast != INSTANT && geo.getDelay() > 0 && !jump) {
				nextColumn = toStack;
				animateSlide(currentColumn, currentDepth,
					nextColumn, nextDepth, fast);
				virtualMove(currentColumn, currentDepth,
					nextColumn, nextDepth);
				currentColumn = nextColumn;
			} else {
				while (currentColumn != toStack) {
					if (jump)
						nextColumn = toStack;
					else
						nextColumn = (currentColumn < toStack) ?
							currentColumn + 1 : currentColumn - 1;
					nextDepth = currentDepth;
					discreteMoves(currentColumn, currentDepth,
						nextColumn, nextDepth);
					currentColumn = nextColumn;
				}
			}
			if (((PanexFrame) frame).getToggleSound()) {
				if (jump)
					((PanexFrame) frame).playDripAudio();
				else
					((PanexFrame) frame).playMoveAudio();
			}
		}
		if (currentDepth < toPosition) {
			if (fast != INSTANT && geo.getDelay() > 0) {
				nextDepth = toPosition;
				animateSlide(currentColumn, currentDepth,
					nextColumn, nextDepth, fast);
				virtualMove(currentColumn, currentDepth,
					nextColumn, nextDepth);
				//currentPosition = nextPosition;
			} else {
				while (currentDepth < toPosition) {
					nextDepth = currentDepth + 1;
					discreteMoves(currentColumn, currentDepth,
						nextColumn, nextDepth);
					currentDepth = nextDepth;
				}
			}
			if (((PanexFrame) frame).getToggleSound())
				((PanexFrame) frame).playMoveAudio();
		}
	}

	int moveTile(int fromStack, int fromPosition, int toStack, int fast) {
		int toPosition;

		if ((toPosition = requestMove(fromStack, fromPosition, toStack,
				false)) >= 0) {
			slideTile(fromStack, fromPosition,
				toStack, toPosition,
				jumpMove(fromStack, toStack), fast);
			((PanexFrame) frame).callback(
				PanexInterface.ACTION_MOVED);
			undo.setMove(fromStack, toStack);
			redo.flushMoves();
		}
		return toPosition;
	}

	int moveTileSilent(int fromStack, int fromPosition, int toStack, int fast) {
		int toPosition;

		if ((toPosition = requestMove(fromStack, fromPosition, toStack,
				false)) >= 0) {
			slideTile(fromStack, fromPosition,
				toStack, toPosition,
				jumpMove(fromStack, toStack), fast);
		}
		return toPosition;
	}

	boolean moveTileToLimbo(int fromStack, int fromPosition, int toStack,
			boolean limbo) {
		boolean aMove;

		aMove = (requestMove(fromStack, fromPosition, toStack,
			limbo) >= 0);
		if (aMove) {
			slideTile(fromStack, fromPosition, toStack, 0,
				jumpMove(fromStack, toStack), NORMAL);
		}
		return aMove;
	}

	void resetTiles() {
		int stack, loc;
		Integer x;

		solve.setFlag(false);
		// setup new Panex with start position set
		model = new PanexFormat(model.getMode(), model.getTiles());
		resizePuzzle();
		startLoc = new Point[model.getColors()][model.getTiles()];
		for (stack = 0; stack < model.getColors(); stack++) {
			x = 2 * stack;
			for (loc = 0; loc < model.getTiles(); loc++) {
				startLoc[stack][loc] =
					new Point(x, loc + model.getBlanks());
			}
		}
		/*positionOfTile = new Point[model.getColors()][model.getTiles()];
		for (stack = 0; stack < model.getColors(); stack++) {
			x = 2 * stack;
			for (loc = 0; loc < model.getTiles(); loc++) {
				positionOfTile[stack][loc] =
					new Point(x, loc + model.getBlanks());
			}
		}
		tileOfPosition = new Point[PanexFormat.MAX_COLUMNS][model.getSlots()];
		for (stack = 0; stack < PanexFormat.MAX_COLUMNS; stack++) {
			for (loc = 0; loc < model.getBlanks(); loc++) {
				tileOfPosition[stack][loc] = new Point(-1, -1);
			}
			if ((x = model.getStartPosition(stack)) >= 0) {
				for (loc = 0; loc < model.getTiles(); loc++) {
					tileOfPosition[stack][loc + model.getBlanks()] =
						new Point(x, loc);
				}
			} else {
				for (loc = 0; loc < model.getTiles(); loc++) {
					tileOfPosition[stack][loc + model.getBlanks()] =
						new Point(-1, -1);
				}
			}

		}*/
		undo.flushMoves();
		redo.flushMoves();
		getStartPosition();
		currentStack = -1;
		started = false;
		cheat = false;
	}

	void moveNoTiles() {
		((PanexFrame) frame).callback(PanexInterface.ACTION_IGNORE);
	}

	int findTile(int positionX) {
		int x = positionX - geo.getFrameOffset().x;
		int i = (x - geo.getDelta().x / 2) / geo.getPos().x;

		if (i < 0)
			i = 0;
		else if (i >= PanexFormat.MAX_COLUMNS)
			i = PanexFormat.MAX_COLUMNS - 1;
		/*
		y -= geo.getFrameOffset().y;
		j = (y - geo.getDelta().y / 2) / geo.getPos().y;
		if (j < 0)
			j = 0;
		else if (j > model.getTiles())
			j = model.getTiles();
		*/
		return i;
	}

	void selectTiles() {
		if ((nextPosition = moveTileSilent(currentStack,
				currentPosition, nextStack, NORMAL)) >= 0) {
			if (previousStack != nextStack) {
				started = true;
				((PanexFrame) frame).callback(
					PanexInterface.ACTION_MOVED);
				if (previousStack >= 0) {
					undo.setMove(previousStack, nextStack);
				} else {
					undo.setMove(currentStack, nextStack);
				}
				redo.flushMoves();
				if (checkSolved()) {
					((PanexFrame) frame).callback(
						PanexInterface.ACTION_SOLVED);
				}
			}
		} else if (nextPosition == -1) {
			((PanexFrame) frame).callback(
				PanexInterface.ACTION_BLOCKED);
			if ((previousStack != -1) &&
					((nextPosition = moveTileSilent(currentStack,
					currentPosition, currentStack, NORMAL)) >= 0) &&
					(previousStack != nextStack) &&
					(previousStack != currentStack)) {
				((PanexFrame) frame).callback(
					PanexInterface.ACTION_MOVED);
				if (previousStack >= 0) {
					undo.setMove(previousStack,
						nextStack);
				} else {
					undo.setMove(currentStack,
						nextStack);
				}
				redo.flushMoves();
				if (checkSolved()) {
					((PanexFrame) frame).callback(
						PanexInterface.ACTION_SOLVED);
				}
			}
		} else {
			((PanexFrame) frame).callback(
				PanexInterface.ACTION_ILLEGAL);
			if (previousStack != -1) {
				nextPosition = moveTileSilent(currentStack,
					currentPosition, previousStack, NORMAL);
			}
		}
		previousStack = -1;
		currentStack = -1;
	}

	public boolean getCheat() {
		return cheat;
	}

	public boolean getStarted() {
		return started;
	}

	public boolean setPuzzleSize(int newValue) {
		if (model.getTiles() != newValue && newValue >= PanexFormat.MIN_TILES) {
			model = new PanexFormat(model.getMode(), newValue);
			geo = null;
			initializePuzzle();
			resizePuzzle();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	public boolean setPuzzleMode(int newValue) {
		if (model.getMode() != newValue &&
				newValue >= PanexFormat.Modes.HANOI.ordinal() &&
				newValue <= PanexFormat.Modes.PANEX.ordinal()) {
			model = new PanexFormat(newValue, model.getTiles());
			geo = null;
			/*if (((PanexFrame) frame).getToggleSound())
				((PanexFrame) frame).playDripAudio();*/
			initializePuzzle();
			resizePuzzle();
			resizePaint = true;
			repaint();
			((PanexFrame) frame).callback(PanexInterface.ACTION_RESET);
			return true;
		}
		return false;
	}

	public int getDelay() {
		return geo.getDelay();
	}

	public void setDelay(int newValue) {
		geo.setDelay(newValue);
	}

	public void speedUpPuzzle() {
		geo.decrementDelay();
		((PanexFrame) frame).callback(PanexInterface.ACTION_SPEED);
	}

	public void slowDownPuzzle() {
		geo.incrementDelay();
		((PanexFrame) frame).callback(PanexInterface.ACTION_SPEED);
	}

	void showMessage(String msg) {
		((PanexFrame) frame).showMessage(msg);
	}

	void resizePuzzle() {
		geo.resize(getSize().width, getSize().height);
		aDraw = new PanexDraw(model, geo, this);
		aDraw.setBuffers(this);
	}

	public void initializePuzzle() {
		if (geo == null) {
			geo = new PanexGeometry(model);
			geo.resize(getWidth(), getHeight());
			aDraw = null;
		} else if (geo.getCoreSize().x != getWidth()
				|| geo.getCoreSize().y != getHeight()) {
			geo.resize(getWidth(), getHeight());
			aDraw = null;
		}
		if (aDraw == null) {
			aDraw = new PanexDraw(model, geo, this);
		}
		checkTiles();
		undo = new PanexStack();
		redo = new PanexStack();
		solve = new PanexSolve(frame, this);
		resetTiles();
	}

	void exposePuzzle() {
		aDraw.getBuffers();
		aDraw.drawFrame(false, false);
		aDraw.drawSlots();
		aDraw.drawAllTiles();
		solve.init();
	}

	void selectPuzzle(int x) {
		int stack;

		previousStack = -1;
		if (nextStack >= 0)
			return;
		if (checkSolved()) {
			moveNoTiles();
		} else if (((stack = findTile(x)) >= 0) &&
				((currentPosition = topOfStack(stack, 0)) >= 0)) {
			currentStack = stack;
			selectPaint = true;
			repaint();
		} else {
			currentStack = -1;
		}
	}

	void motionPuzzle(int x) {
		int stack;

		if (currentStack < 0 || selectPaint)
			return;
		if (checkSolved()) {
			moveNoTiles();
		} else if (((stack = findTile(x)) >= 0) &&
				stack != currentStack) {
			nextStack = stack;
			motionPaint = true;
			repaint();
		}
	}

	void releasePuzzle(int x) {
		if (currentStack < 0)
			return;
		nextStack = findTile(x);
		releasePaint = true;
		repaint();
	}

	public void getPuzzle() {
		this.stringSave = ((PanexFrame) frame).readTextArea();
		if (this.stringSave == null || this.stringSave.equals("")) {
			showMessage("Empty string to read");
			return;
		}
		getPuzzleString("buffer");
	}

	public void getPuzzleFile() {
		undo.flushMoves();
		redo.flushMoves();
		getStartPosition();
		try (BufferedReader br = new BufferedReader(
				new FileReader(DATAFILE))) {
			StringBuilder puzzleData = new StringBuilder();
			String lineData;

			while ((lineData = br.readLine()) != null)
				puzzleData.append(lineData).append(NL);
			this.stringSave = puzzleData.toString();
			if (this.stringSave == null || this.stringSave.equals("")) {
				showMessage("Empty file " + DATAFILE);
				return;
			}
			getPuzzleString(DATAFILE);
		} catch (IOException e) {
			//e.printStackTrace();
			showMessage("Cannot read from " + DATAFILE);
		}
	}

	public void getPuzzleString(String source) {
		try {
			int i, tilesValue, modeValue, moves;
			String symbol = SYMBOL;
			String newline = NL;
			int line = 0;

			token = this.stringSave.split(newline);
			String lineData = token[line];
			if (lineData == null || lineData.length() == 0) {
				showMessage(source +
					" nothing in buffer");
				return;
			}
			tilesValue = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			if (tilesValue >= PanexFormat.MIN_TILES) {
				for (i = this.model.getTiles(); i < tilesValue; i++) {
					((PanexFrame) frame).callback(
						PanexInterface.ACTION_INCREMENT);
				}
				for (i = this.model.getTiles(); i > tilesValue; i--) {
					((PanexFrame) frame).callback(
						PanexInterface.ACTION_DECREMENT);
				}
			} else {
				showMessage(source +
					" corrupted: tiles " + tilesValue +
					" should be at least " +
					PanexFormat.MIN_TILES);
				return;
			}
			lineData = token[++line];
			modeValue = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			if (modeValue < PanexFormat.Modes.HANOI.ordinal() ||
					modeValue > PanexFormat.Modes.PANEX.ordinal())
				modeValue = PanexFormat.defaultMode.ordinal();
			if (model.getMode() != modeValue)
				((PanexFrame) frame).callback(
					PanexInterface.ACTION_MODE);
			if (model.getMode() != modeValue)
				((PanexFrame) frame).callback(
					PanexInterface.ACTION_MODE);
			lineData = token[++line];
			moves = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			line = scanStartPosition(++line);
			if (line < 0)
				return;
			((PanexFrame) frame).callback(
				PanexInterface.ACTION_RESTORE);
			model.setStartPosition();
			resetPaint = true;
			repaint();
			if (!scanMoves(++line, moves))
				return;
			System.out.println(source + ": tiles " + tilesValue +
				", mode " + modeValue + ", moves " + moves);
			cheat = true;
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			showMessage("Corrupt input found when reading");
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Bad input file");
		}
	}

	public void printPuzzle(StringBuilder sb) {
		String symbol = SYMBOL;
		String space = " ";
		String newline = NL;

		sb.append("tiles").append(symbol).append(space)
			.append(model.getTiles()).append(newline);
		sb.append("mode").append(symbol).append(space)
			.append(model.getMode()).append(newline);
		sb.append("moves").append(symbol).append(space)
			.append(undo.numMoves()).append(newline);
		printStartPosition(sb);
		printMoves(sb);
	}

	public void writePuzzle() {
		StringBuilder sb = new StringBuilder();

		printPuzzle(sb);
		((PanexFrame) frame).writeTextArea(sb.toString());
	}

	public void writePuzzleFile() {
		StringBuilder sb = new StringBuilder();

		printPuzzle(sb);
		try (PrintWriter pr = new PrintWriter(new FileWriter(
				DATAFILE, false))) {
			pr.print(sb);
			System.out.println("Saved to " + DATAFILE);
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Cannot write to " + DATAFILE);
		}
	}

	public void clearPuzzle() {
		if (mouseDown)
			return;
		resetPaint = true;
		repaint();
		((PanexFrame) frame).callback(PanexInterface.ACTION_RESET);
	}

	public void clearWithQueryPuzzle() {
		if (!started)
			clearPuzzle();
	}

	public void undoPuzzle() {
		if (mouseDown || stackPaint || !undo.madeMoves())
			return;
		undo.getMove();
		// move in opposite direction
		nextStack = undo.readMoveFrom();
		currentStack = undo.readMoveTo();
		redo.setMove(nextStack, currentStack);
		if ((currentPosition = topOfStack(currentStack, 0)) < 0 ||
				requestMove(currentStack, currentPosition,
				nextStack, false) < 0) {
			currentStack = -1;
		} else {
			stackPaint = true;
			repaint();
			started = true;
			((PanexFrame) frame).callback(
				PanexInterface.ACTION_UNDO);
		}
	}

	public void redoPuzzle() {
		if (mouseDown || stackPaint || !redo.madeMoves())
			return;
		redo.getMove();
		currentStack = redo.readMoveFrom();
		nextStack = redo.readMoveTo();
		undo.setMove(currentStack, nextStack);
		if ((currentPosition = topOfStack(currentStack, 0)) < 0 ||
				requestMove(currentStack, currentPosition,
				nextStack, false) < 0) {
			currentStack = -1;
		} else {
			stackPaint = true;
			repaint();
			started = true;
			((PanexFrame) frame).callback(
				PanexInterface.ACTION_REDO);
		}
	}

	public void solvePuzzle() {
		if (mouseDown || stackPaint)
			return;
		/* Cheat a little and reset to start. */
		Graphics g = getGraphics();

		clearPuzzle();
		try {
			paint(g);
		} catch (Exception e) {
			e.printStackTrace();
			return;
		} finally {
			g.dispose();
		}
		cheat = true;
		solve.setFlag(true);
		try {
			Thread.sleep(200);
		} catch (InterruptedException e) {
			System.err.println("Interrupted");
		}
		//panexSolve.solveTiles();
	}

	public void incrementPuzzle() {
		if (mouseDown)
			return;
		((PanexFrame) frame).callback(PanexInterface.ACTION_INCREMENT);
	}

	public void decrementPuzzle() {
		if (mouseDown || model.getTiles() <= PanexFormat.MIN_TILES)
			return;
		((PanexFrame) frame).callback(PanexInterface.ACTION_DECREMENT);
	}

	public void modePuzzle() {
		if (mouseDown)
			return;
		((PanexFrame) frame).callback(PanexInterface.ACTION_MODE);
	}

	void getStartPosition() {
		for (int color = 0; color < model.getColors(); color++)
			for (int number = 0; number < model.getTiles(); number++) {
				startLoc[color][number].x =
					model.getColumnOfTile(color, number);
				startLoc[color][number].y =
					model.getDepthOfTile(color, number);
			}
	}

	int scanStartPosition(int location) {
		String lineData;
		int line = location;

		try {
			String space = "\\s+";

			++line; //read startingPosition:
			for (int position = 0; position < model.getTiles(); position++) {
				line++;
				if (line >= token.length) {
					System.out.println("Short data in StartPosition");
					return -1;
				}
				lineData = token[line];
				String[] st = lineData.split(space);
				int offset = ("".equals(st[0])) ? 1 : 0;
				for (int stack = 0; stack < model.getColors(); stack++) {
					startLoc[stack][position].x =
						Integer.parseInt(st[2 * stack + offset].trim());
					startLoc[stack][position].y =
						Integer.parseInt(st[2 * stack + offset + 1].trim());
				}
			}
			//blank line read
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			showMessage("Corrupt input in StartPosition");
			return -1;
		}
		return line;
	}

	boolean scanMoves(int location, int moves) {
		Graphics g = getGraphics();

		if (g == null)
			return false;
		try {
			int l;
			int line = location;
			String lineData;
			String space = "\\s+";

			++line; //read moves direction
			paint(g);
			for (l = 0; l < moves && ++line < token.length; l++) {
				lineData = token[line];
				String[] st = lineData.split(space);
				currentStack = Integer.parseInt(st[1].trim());
				nextStack = Integer.parseInt(st[2].trim());
				if ((currentPosition =
						topOfStack(currentStack, 0)) < 0 ||
						requestMove(currentStack,
							currentPosition, nextStack,
							false) < 0) {
					System.out.println(l +
						": move from " +
						currentStack + " to " +
						nextStack +
						", cannot be made.");
					currentStack = -1;
					return false;
				} else {
					scanPaint = true;
					undo.setMove(currentStack, nextStack);
					redo.flushMoves();
					paint(g);
					started = true;
					((PanexFrame) frame).callback(
						PanexInterface.ACTION_MOVED);
				}
			}
			if (l < moves) {
				showMessage("Corrupt input in scanMoves, only " + l + " moves");
				return false;
			}
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			showMessage("Corrupt input in scanMoves");
			return false;
		} finally {
			g.dispose();
		}
		return true;
	}

	void printStartPosition(StringBuilder sb) {
		String symbol = SYMBOL;
		String space = " ";
		String newline = NL;

		sb.append(newline);
		sb.append("startingPosition").append(symbol).append(newline);
		for (int position = 0; position < model.getTiles(); position++)
			for (int stack = 0; stack < model.getColors(); stack++) {
				int location = startLoc[stack][position].x;
				if (location < BASE * BASE)
					sb.append(space);
				if (location < BASE)
					sb.append(space);
				sb.append(space).append(location);
				location = startLoc[stack][position].y;
				if (location < BASE)
					sb.append(space);
				sb.append(space).append(location);
				if (stack == model.getColors() - 1)
					sb.append(newline);
			}
		sb.append(newline);
	}

	void printMoves(StringBuilder sb) {
		int from, to;
		int nMoves = undo.numMoves();
		String symbol = SYMBOL;
		String space = "\t";
		String newline = NL;

		sb.append("moves").append(space)
			.append("from").append(space)
			.append("to").append(newline);
		while (undo.madeMoves()) {
			undo.getMove();
			from = undo.readMoveFrom();
			to = undo.readMoveTo();
			redo.setMove(from, to);
		}
		// apply all moves back to the stack since
		// we emptied it above!!
		for (int i = 0; i < nMoves; i++) {
			redo.getMove();
			from = redo.readMoveFrom();
			to = redo.readMoveTo();
			sb.append(i + 1).append(symbol).append(space)
				.append(from).append(space)
				.append(to).append(newline);
			undo.setMove(from, to);
		}
	}

	@Override
	public void paint(Graphics g) {
		if (firstPaint) {
			if (model == null)
				return;
			initializePuzzle();
			firstPaint = false;
			resizePaint = true;
		}
		if (!framePaint && !resetPaint &&
				!selectPaint && !releasePaint &&
				!motionPaint && !movePaint && !scanPaint &&
				!stackPaint)
			resizePaint = true;
		if (resizePaint) {
			resizePuzzle();
			exposePuzzle();
			resizePaint = false;
		}
		if (framePaint) {
			aDraw.drawFrame(true, true);
			framePaint = false;
		}
		if (resetPaint) {
			resetTiles();
			aDraw.getBuffers();
			aDraw.drawFrame(true, false);
			aDraw.drawSlots();
			aDraw.drawAllTiles();
			resetPaint = false;
		}
		if (selectPaint) {
			if (moveTileToLimbo(currentStack, currentPosition,
					currentStack, false)) {
				currentPosition = 0;
				previousStack = currentStack;
			} else if (moveTileToLimbo(currentStack, currentPosition,
					currentStack, true)) {
				currentPosition = 0;
			}
			aDraw.drawTile(currentStack, currentPosition, false, 1, 0, 0);
			selectPaint = false;
		}
		if (motionPaint) {
			if (moveTileToLimbo(currentStack, currentPosition,
					nextStack, true)) {
				if (previousStack < 0) {
					previousStack = currentStack;
				}
				currentStack = nextStack;
				currentPosition = 0;
				aDraw.drawTile(currentStack, currentPosition, true,
					1, 0, 0);
				aDraw.drawTile(nextStack, currentPosition, false,
					1, 0, 0);
			} else if (moveTileToLimbo(currentStack, currentPosition,
					nextStack, false)) {
				currentStack = nextStack;
				currentPosition = 0;
				aDraw.drawTile(currentStack, currentPosition, true,
					1, 0, 0);
				aDraw.drawTile(nextStack, currentPosition, false,
					1, 0, 0);
			}
			motionPaint = false;
		}
		if (releasePaint) {
			aDraw.drawTile(currentStack, currentPosition, true, 1, 0, 0);
			aDraw.drawTile(currentStack, currentPosition, false, 0, 0, 0);
			if (nextStack >= 0 &&
					((previousStack < 0 && nextStack != currentStack) ||
					previousStack >= 0)) {
				selectTiles();
			}
			releasePaint = false;
			currentStack = -1;
			previousStack = -1;
			nextStack = -1;
		}
		if (movePaint || scanPaint || stackPaint) {
			int fast;

			if (stackPaint)
				fast = DOUBLE;
			else if (movePaint)
				fast = NORMAL;
			else
				fast = INSTANT;
			/* nextPosition = */ moveTileSilent(currentStack,
				currentPosition, nextStack, fast);
			currentStack = -1;
			nextStack = -1;
			movePaint = false;
			scanPaint = false;
			stackPaint = false;
			try {
				Thread.sleep(geo.getDelay());
			} catch (InterruptedException e) {
				// ignore
			}
		}
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}
}
